\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\ GNU General Public License for more details.

\ You should have received a copy of the GNU General Public License
\ along with this program; if not, write to the Free Software
\ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

require lib.fs
require string.fs
require bitarray.fs
require tasker.fs
require unix/socket.fs

libc bind int ptr ptr (int) bind ( sockfd my_addr socklen_t -- err)
libc listen int int (int) listen ( sockfd maxnum -- err ) 
libc accept_socket int int int (int) accept ( sockfd cliadr adrlen -- err ) 
libc shutdown_socket int int (int) shutdown ( sockfd method -- err ) 
libc close_socket int (int) close ( sockfd -- ) 
libc recv_socket int ptr int int (int) recv ( sockfd buff len flags -- ) 
libc fcntl int int int (int) fcntl ( sockfd cmd argi -- res)
libc send_socket int ptr int int (int) send ( sockfd ptr len flags -- err )
libc select int int int int  ptr (int) select ( sockfd sockfd sockfd sockfd timeout -- res )


20000 Constant MAX_CONNECTIONS
0 Constant NULL
64 Constant MSG_WAITALL 
3 Constant F_GETFL
4 Constant F_SETFL
128 Constant O_NONBLOCK


: new-socket-bind ( -- socket )
    PF_INET SOCK_STREAM 0 socket
    dup 0<= abort" no free bind socket" ;

: >inetaddr-bind ( ip port sockaddr -- ) >r
    r@ sockaddr_in %size erase
    PF_INET r@ family w!
    htons r@ port w!
    htonl r> sin_addr l! ;

: open-socket-bind ( addr u port -- fid )
    -rot host>addr
    swap sockaddr-tmp >inetaddr
    new-socket-bind \ fid now on stack
    dup sockaddr-tmp sockaddr_in %size bind 0< abort" could not bind to socket" \ still on stack
    dup MAX_CONNECTIONS listen 0< abort" will not listen to socket"
    cr ." Now listening to socket, hey!" cr
;
    
: close-the-socket ( fid -- )
    try
	close_socket
    restore
	\ always fails for some reason?
    endtry
    throw 
;


struct
   cell% field sd_inbuffer
   cell% field sd_outbuffer
   cell% field sd_inbuffer_size
   cell% field sd_outbuffer_size
   cell% field sd_received_data_length
   cell% field sd_client_data
   cell% field sd_time_struct
   cell% field sd_callback_response
   cell% field sd_callback_accept
end-struct server-data


: multi-server
	64 postpone literal
	postpone NewTask
	2 postpone literal
	postpone swap
	postpone pass ; immediate
	


256 Constant max-line
Create line-buffer  max-line 2 + allot

: set-socket-nonblock ( fid -- fid )
    dup F_GETFL 0 fcntl 128 or 128 xor { sock_flag }
    dup F_SETFL sock_flag fcntl drop \ ." setfl to: " . cr 
;
	
\ 
\  This is the callback to handle incoming data. To make it multitasking include multi-server as the first line.  
\  It is supposed to work like this:
\	- Call multitasker
\	- Get data from input-buffer or/and read from sd (server-data-struct). This content mutates after a call to pause, so all needed data after a pause must be copied, as all the connections share a common sd-struct.
\	- Do the calculations or whatever. Loops should contain the word pause, to keep the multitasker going
\	- Send data back to client
\	- Close connection or leave this to the client. Use close-the-server if you choose to close.

: send-socket-line ( fid sd -- )
	multi-server
	{ sd }
	sd sd_inbuffer @ sd sd_received_data_length @ type
	dup s\" Hello Major Tom. Are you receiving?\n" 0 send_socket 
        dup 0< if abort" Count not send line" endif \ fix, check that all data was sent
	2drop
;

: read-socket-line ( fid sd -- res )
   over over sd_inbuffer @ over sd_inbuffer_size 0 recv_socket
   dup 0< if abort" Could not readline" endif
   swap drop
; 


\ override this callback in default_server_data-struct to change the procedure to something sensible.
: socket-accepted ( fid sd -- )
   multi-server
   { fd }
  ." Socket accepted on port: " . cr 
  ." Connection from: " fd sd_client_data @ sin_addr @ hex . decimal cr  
;

0 Value connection-counter
create time_struct 0 , 0 ,

: create_default_server_data
	server-data %allot { sd }
	line-buffer sd sd_inbuffer !
	line-buffer sd sd_outbuffer !
	max-line sd sd_inbuffer_size !
	max-line sd sd_outbuffer_size !
	0 sd sd_received_data_length !
	0 sd sd_client_data !
	time_struct sd sd_time_struct !
	['] send-socket-line sd sd_callback_response !
	['] socket-accepted sd sd_callback_accept !
	sd
;

: open-socket-accept ( sd port -- fid )
    swap { sd }
    new_fd_struct new_fd_struct  { fd_read fd_write }
    s" localhost" rot open-socket-bind dup { bindPort } \ fid 
    set-socket-nonblock \ fid
    ." Waiting for new connections" cr .s cr
    fd_read over FD_SET
    dup { max_fd }
    sockaddr_in %allot { sockaddr } sockaddr sockaddr_in %size erase
    sockaddr sd sd_client_data !
    here { sock_length } cell allot

    begin
        fd_read fd_write FD_CPY \ bind-fid 
	max_fd 1+ fd_write NULL NULL sd sd_time_struct @ select
	dup 0< if abort" error in select" endif
	0> if 
		max_fd 1+ 0 +do
		   fd_write i FD_GET if
			   i bindPort = if
			      sockaddr_in %size sock_length !
			      dup sockaddr sock_length accept_socket dup 0< abort" Socket not accepted"  
			      set-socket-nonblock
			      dup max_fd max to max_fd \ listenfid newfid
			      dup fd_read swap FD_SET  
			      sd sd sd_callback_accept @ execute
			   else
			      i sd read-socket-line 
			      dup 0= if
				2drop i close-the-socket drop
				fd_read i FD_NOT bindport
			      else \ fid read-length
				sd sd_received_data_length !
				sd sd sd_callback_response @ execute \ fid
			      endif		 
			   endif
		   endif
		loop
	else 
		\ select has timed out	
        endif
	connection-counter 1+ to connection-counter
	\ connection-counter 5 > if abort" End" endif
	pause
	\ connection-counter . cr
    again
;

